import math


def wrap_at_limit(text: str, character_limit: int, include_existing_linebreaks: bool = True) -> str:
    """Wrap the given text to the given number of characters.  May split long words in the middle."""
    pass


def wrap_at_square(text: str) -> str:
    """Try as best as possible to make the text into a square."""
    pass